
/*
** Copyright (c) 2006-2007 Herbert Poetzl

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
Version 2 as published by the Free Software Foundation.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.


** V0.01	initial prototype
** V0.02	update to 2.1.1 bcaps
** V0.03	added numeric cmd parsing
**		added compat support
** V0.04	added missing task vcmds
** V0.05	fixed network byteorder
**		added pcap support
** V0.06	posix capset support
** V0.07	basic net and device support
** V0.08	clone and unshare support
** V0.09	fixed clone and messages
** V0.10	added sched v5 support
** V0.11	network v1 support
** V0.12	license for packaging
** V0.13	added tag cmds
**
*/

#define VERSION "V0.13"

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdint.h>
#include <netinet/in.h>
#include <linux/capability.h>

#include "vserver.h"
#include "clone.h"
#include "unshare.h"
#include "switch.h"

#include "network.h"
#include "legacy.h"

#include "context_cmd.h"
#include "debug_cmd.h"
#include "inode_cmd.h"
#include "space_cmd.h"
#include "sched_cmd.h"
#include "cvirt_cmd.h"
#include "dlimit_cmd.h"
#include "limit_cmd.h"
#include "network_cmd.h"
#include "signal_cmd.h"
#include "device_cmd.h"
#include "tag_cmd.h"

static	char	opt_id = 0;
static	char	opt_cmd = 0;
static	char	opt_cap = 0;
static	char	opt_vers = 0;
static	char	opt_desc = 0;
static	char	opt_cont = 0;
static	char	opt_list = 0;
static	char	opt_mask = 0;
static	char	opt_quiet = 0;
static	char	opt_terse = 0;
static	char	opt_after = 0;
static	char	opt_before = 0;
static	char	opt_dryrun = 0;
static	char	opt_clone = 0;
static	char	opt_unshare = 0;

static	char *	cmd_name = NULL;
static	char *	exec_name = NULL;

static uint32_t	num_id = 0;
static uint32_t	num_ret = 0;
static uint32_t	num_cmd = 0;
static uint32_t	num_mask = 0;

static int	num_err = 0;

static uint32_t	cap_eff = 0;
static uint32_t	cap_prm = 0;
static uint32_t	cap_inh = 0;

static uint64_t	flags_clone = 0;
static uint64_t	flags_unshare = 0;


typedef long long int (stoll_t)(const char *, char **, int);

long long int argtoll(const char *str, stoll_t stoll)
{
	int bit, inv = 0, nbo = 0;
	long long int val = 0;
	char *eptr;

	if (!str)
	    return -1;
	if (!stoll)
	    stoll = strtoll;
	
	switch (*str) {
	case '~':
	case '!':
	    inv = 1;	/* invert */
	    str++;
	default:
	    break;
	}

	while (*str) {
	    switch (*str) {
	    case '^':
		bit = strtol(str+1, &eptr, 0);
		val ^= (1LL << bit);
		break;
	    case '&':
		val &= stoll(str+1, &eptr, 0);
		break;
	    case '|':
		val |= stoll(str+1, &eptr, 0);
		break;
	    case '.':
		val = (val << 8) | (strtol(str+1, &eptr, 0) & 0xff);
		nbo++;
		break;
	    default:
		val = stoll(str, &eptr, 0);
		break;
	    }
	    if (eptr == str)
	    	break;
	    str = eptr;
	}
	if (nbo>2)
	    val = htonl(val);
	else if (nbo>0)
	    val = htons(val);
	return (inv)?~(val):(val);
}


#define	ATO_TYPE(_type)			\
_type ato_##_type (const char *str) {	\
	_type val = argtoll(str,0);	\
	return val;			\
}

ATO_TYPE(uint64_t)
#define	FMT_uint64_t	"0x%016llX (%llu)"

ATO_TYPE(uint32_t)
#define	FMT_uint32_t	"0x%08llX (%llu)"

ATO_TYPE(uint16_t)
#define	FMT_uint16_t	"0x%04llX (%llu)"

ATO_TYPE(int64_t)
#define	FMT_int64_t	"0x%016llX (%lld)"

ATO_TYPE(int32_t)
#define	FMT_int32_t	"0x%08llX (%lld)"

ATO_TYPE(int16_t)
#define	FMT_int16_t	"0x%04llX (%lld)"


#define	FMT_INDEX	"  [%d] .%-16s"


#define NIPQUAD(addr) \
	((unsigned char *)&addr)[0], \
	((unsigned char *)&addr)[1], \
	((unsigned char *)&addr)[2], \
	((unsigned char *)&addr)[3]
#define NIPQUAD_FMT "%u.%u.%u.%u"

#define NIP6(addr) \
	ntohs((addr).s6_addr16[0]), \
	ntohs((addr).s6_addr16[1]), \
	ntohs((addr).s6_addr16[2]), \
	ntohs((addr).s6_addr16[3]), \
	ntohs((addr).s6_addr16[4]), \
	ntohs((addr).s6_addr16[5]), \
	ntohs((addr).s6_addr16[6]), \
	ntohs((addr).s6_addr16[7])
#define NIP6_FMT "%04x:%04x:%04x:%04x:%04x:%04x:%04x:%04x"


void ato_in_addr (const char *str, struct in_addr *val) {
	val->s_addr = argtoll(str,0);
}


void out_in_addr (int index, const char *name, struct in_addr *val) {
	printf(FMT_INDEX " " NIPQUAD_FMT, index, name,
		NIPQUAD(*val));
}

void ato_in6_addr (const char *str, struct in6_addr *val) {
	val->s6_addr32[0] = argtoll(str,0);
}

void out_in6_addr (int index, const char *name, struct in6_addr *val) {
	printf(FMT_INDEX " " NIP6_FMT, index, name,
		NIP6(*val));
}


#define	EENTRY(_symb,_desc) {	\
	.code = _symb,		\
	.symb = #_symb,		\
	.desc = _desc,		\
	}

struct	err_data {
	int code;
	char *symb;
	char *desc;
} err_data[] = {
	EENTRY(EPERM,	"Insufficient privileges"),
	EENTRY(ENOENT,	"No such entity"),
	EENTRY(ESRCH,	"No such process or context"),
	EENTRY(EINTR,	"Interrupted system call"),
	EENTRY(ENXIO,	"No such device or address"),
	EENTRY(EAGAIN,	"Try again"),
	EENTRY(ENOMEM,	"Out of memeory"),
	EENTRY(EACCES,	"Permission denied"),
	EENTRY(EFAULT,	"Bad address"),
	EENTRY(ENOTTY,	"Not A Typewriter :)"),
	EENTRY(ENOSYS,	"Function not implemented"),
	EENTRY(EINVAL,	"Invalid argument or data"),
	EENTRY(0,	"Success"),
	EENTRY(-1,	"Unknown error"),
};

#define	CENTRY(_cap) {		\
	.bit = CAP_##_cap,	\
	.name = #_cap,		\
	}

struct	bit_data {
	int bit;
	char *name;
} bcap_data[] = {
	CENTRY(CHOWN),
	CENTRY(DAC_OVERRIDE),
	CENTRY(DAC_READ_SEARCH),
	CENTRY(FOWNER),
	CENTRY(FSETID),
	CENTRY(KILL),
	CENTRY(SETGID),
	CENTRY(SETUID),
	CENTRY(SETPCAP),
	CENTRY(LINUX_IMMUTABLE),
	CENTRY(NET_BIND_SERVICE),
	CENTRY(NET_BROADCAST),
	CENTRY(NET_ADMIN),
	CENTRY(NET_RAW),
	CENTRY(IPC_LOCK),
	CENTRY(IPC_OWNER),
	CENTRY(SYS_MODULE),
	CENTRY(SYS_RAWIO),
	CENTRY(SYS_CHROOT),
	CENTRY(SYS_PTRACE),
	CENTRY(SYS_PACCT),
	CENTRY(SYS_ADMIN),
	CENTRY(SYS_BOOT),
	CENTRY(SYS_NICE),
	CENTRY(SYS_RESOURCE),
	CENTRY(SYS_TIME),
	CENTRY(SYS_TTY_CONFIG),
	CENTRY(MKNOD),
	CENTRY(LEASE),
#ifdef	CAP_AUDIT_WRITE
	CENTRY(AUDIT_WRITE),
	CENTRY(AUDIT_CONTROL),
#endif
#ifdef	CAP_CONTEXT
	CENTRY(CONTEXT),
#endif
	{-1, NULL}
}, ccap_data[] = {
}, flag_data[] = {
};

int	parse_bit(const char *str, char **endptr, struct bit_data *data)
{
	while (data->name) {
	    if (strcmp(data->name, str) == 0) {
		if (endptr)
		    *endptr = (char *)str + strlen(data->name);
		return data->bit;
	    }
	    data++;
	}
	if (endptr)
	    *endptr = (char *)str;
	return -1;
}


long long int stobcap(const char *str, char **endptr, int base)
{
	unsigned long long ret = 0;
	int bit;

	if (strncmp(str, "CAP_", 4) == 0)
	    str += 4;

	bit = parse_bit(str, endptr, bcap_data);
	ret = (bit<0) ? strtoll(str, endptr, base) : (1LL << bit);
	return ret;
}

void parse_caps(char *str)
{
	char *str_eff, *str_prm, *str_inh;

	str_inh = strtok(str, ":");
	str_prm = strtok(NULL, ":");
	str_eff = strtok(NULL, ":");

	if (str_prm)
		cap_prm = argtoll(str_prm, stobcap);
	if (str_inh)
		cap_inh = argtoll(str_inh, stobcap);
	if (str_eff)
		cap_eff = argtoll(str_eff, stobcap);
}


#define	CMD_COND(_cmd,_index,_out)			\
	((out == _out) && (cmd == _cmd) && (index == _index))


#define	CMD_FILL_T(_cmd,_index,_type,_ent,_name) do {	\
	if (CMD_COND(_cmd,_index,0)) {			\
	    data->_ent._name = ato_##_type(str);	\
	    return 1;					\
	}} while (0)

#define	CMD_FILL_S(_cmd,_index,_len,_ent,_name) do {	\
	if (CMD_COND(_cmd,_index,0)) {			\
	    strncpy(data->_ent._name, str, _len);	\
	    return 1;					\
	}} while (0)

#define	CMD_FILL_P(_cmd,_index,_ent,_name) do {		\
	if (CMD_COND(_cmd,_index,0)) {			\
	    data->_ent._name = str;			\
	    return 1;					\
	}} while (0)

#define	CMD_FILL_N(_cmd,_index,_type,_ent,_name) do {	\
	if (CMD_COND(_cmd,_index,0)) {			\
	    ato_##_type(str, &data->_ent._name);	\
	    return 1;					\
	}} while (0)


#define	CMD_OUT_T(_cmd,_index,_type,_ent,_name) do {	\
	if (CMD_COND(_cmd,_index,1)) {			\
	    printf(FMT_INDEX " " FMT_##_type "\n",	\
		_index, #_name,				\
		(long long)data->_ent._name,		\
		(long long)data->_ent._name);		\
	    return 1;					\
	}} while (0)

#define	CMD_OUT_S(_cmd,_index,_len,_ent,_name) do {	\
	if (CMD_COND(_cmd,_index,1)) {			\
	    printf(FMT_INDEX " %.*s\n", _index,	\
		#_name, _len, data->_ent._name);	\
	    return 1;					\
	}} while (0)

#define	CMD_OUT_P(_cmd,_index,_ent,_name) do {		\
	if (CMD_COND(_cmd,_index,1)) {			\
	    printf(FMT_INDEX " PTR: %p %s\n",		\
		_index, #_name,				\
		data->_ent._name,			\
		data->_ent._name);			\
	    return 1;					\
	}} while (0)

#define	CMD_OUT_N(_cmd,_index,_type,_ent,_name) do {	\
	if (CMD_COND(_cmd,_index,1)) {			\
	    out_##_type(_index, #_name,			\
		&data->_ent._name);			\
	    return 1;					\
	}} while (0)


#define	CMD_ICOND(_cmd) ((out == -1) && (cmd == _cmd))

#define	CMD_INDEX(_cmd,_index,_name) do {		\
	if (CMD_ICOND(_cmd) && (strcmp(str,_name)==0))	\
	    return _index;				\
	} while (0)

#define	CMD_VCMD(_cmd,_name,_index,_io) do {		\
	if ((out == -2) && (strcmp(str,_name)==0))	\
	    return _cmd;				\
	if ((out == -4) && (cmd == _cmd)) {		\
	    data->info.cmd_name = _name;		\
	    return 1;					\
	}						\
	if (CMD_COND(_cmd,_index,-2))			\
	    return _io;					\
	} while (0)

#define	CMD_DESC(_cmd,_name,_index,_io,_type) do {	\
	if (CMD_COND(_cmd,_index,-3)) {			\
	    printf(FMT_INDEX " (%s)\n",			\
		_index, _name,				\
		(_io == 3)?"in/out":			\
		((_io == 1)?"in":"out"));		\
	    return 1;					\
	}} while (0)


#define	CMDE_T(_cmd,_ext,_index,_io,_type,_name) do {		\
	CMD_OUT_T(VCMD_##_cmd,_index,_type,_cmd##_ext,_name);	\
	CMD_FILL_T(VCMD_##_cmd,_index,_type,_cmd##_ext,_name);	\
	CMD_INDEX(VCMD_##_cmd,_index,#_name);			\
	CMD_DESC(VCMD_##_cmd,#_name,_index,_io,#_type);		\
	CMD_VCMD(VCMD_##_cmd,#_cmd,_index,_io);			\
	} while (0)

#define	CMDE_S(_cmd,_ext,_index,_io,_len,_name) do {		\
	CMD_OUT_S(VCMD_##_cmd,_index,_len,_cmd##_ext,_name);	\
	CMD_FILL_S(VCMD_##_cmd,_index,_len,_cmd##_ext,_name);	\
	CMD_INDEX(VCMD_##_cmd,_index,#_name);			\
	CMD_DESC(VCMD_##_cmd,#_name,_index,_io,"str_"#_len);	\
	CMD_VCMD(VCMD_##_cmd,#_cmd,_index,_io);			\
	} while (0)

#define	CMDE_P(_cmd,_ext,_index,_io,_name) do {			\
	CMD_OUT_P(VCMD_##_cmd,_index,_cmd##_ext,_name);		\
	CMD_FILL_P(VCMD_##_cmd,_index,_cmd##_ext,_name);	\
	CMD_INDEX(VCMD_##_cmd,_index, #_name);			\
	CMD_DESC(VCMD_##_cmd,#_name,_index,_io,"ptr");		\
	CMD_VCMD(VCMD_##_cmd,#_cmd,_index,_io);			\
	} while (0)

#define	CMDE_N(_cmd,_ext,_index,_io,_type,_name) do {		\
	CMD_OUT_N(VCMD_##_cmd,_index,_type,_cmd##_ext,_name);	\
	CMD_FILL_N(VCMD_##_cmd,_index,_type,_cmd##_ext,_name);	\
	CMD_INDEX(VCMD_##_cmd,_index,#_name);			\
	CMD_DESC(VCMD_##_cmd,#_name,_index,_io,#_type);		\
	CMD_VCMD(VCMD_##_cmd,#_cmd,_index,_io);			\
	} while (0)

#define	CMDE_X(_cmd) do {					\
	CMD_VCMD(VCMD_##_cmd,#_cmd,-1,0);			\
	} while (0)


struct	meta_info {
	const char *cmd_name;
};

union	cmd_arg {
	struct meta_info		info;

	/* legacy compat */
	struct vcmd_new_s_context_v1	new_s_context_v1;
	struct vcmd_set_ipv4root_v3	set_ipv4root_v3;

	/* from context_cmd */
	struct vcmd_vx_info_v0		vx_info;
	struct vcmd_ctx_create		ctx_create;
	struct vcmd_ctx_migrate		ctx_migrate;
	struct vcmd_ctx_flags_v0	get_cflags;
	struct vcmd_ctx_flags_v0	set_cflags;
	struct vcmd_ctx_caps_v0		get_ccaps_v0;
	struct vcmd_ctx_caps_v0		set_ccaps_v0;
	struct vcmd_ctx_caps_v1		get_ccaps_v1;
	struct vcmd_ctx_caps_v1		set_ccaps_v1;
	struct vcmd_bcaps		get_bcaps;
	struct vcmd_bcaps		set_bcaps;

	/* from network_cmd */
	struct vcmd_nx_info_v0		nx_info;
	struct vcmd_net_create		net_create;
	struct vcmd_net_addr_v0		net_add;
	struct vcmd_net_addr_v0		net_remove;
	struct vcmd_net_addr_ipv4_v1	net_add_ipv4;
	struct vcmd_net_addr_ipv4_v1	net_remove_ipv4;
	struct vcmd_net_addr_ipv6_v1	net_add_ipv6;
	struct vcmd_net_addr_ipv6_v1	net_remove_ipv6;
	struct vcmd_net_flags_v0	get_nflags;
	struct vcmd_net_flags_v0	set_nflags;
	struct vcmd_net_caps_v0		get_ncaps;
	struct vcmd_net_caps_v0		set_ncaps;

	/* from cvirt_cmd */
	struct vcmd_vhi_name_v0		set_vhi_name;
	struct vcmd_vhi_name_v0		get_vhi_name;

	/* from dlimit_cmd */
	struct vcmd_ctx_dlimit_base_v0	add_dlimit;
	struct vcmd_ctx_dlimit_base_v0	rem_dlimit;
	struct vcmd_ctx_dlimit_v0	set_dlimit;
	struct vcmd_ctx_dlimit_v0	get_dlimit;

	/* from inode_cmd */
	struct vcmd_ctx_iattr_v0	get_iattr_v0;
	struct vcmd_ctx_iattr_v0	set_iattr_v0;
	struct vcmd_ctx_iattr_v1	get_iattr_v1;
	struct vcmd_ctx_iattr_v1	set_iattr_v1;

	/* from rlimit_cmd */
	struct vcmd_ctx_rlimit_v0	get_rlimit;
	struct vcmd_ctx_rlimit_v0	set_rlimit;
	struct vcmd_ctx_rlimit_mask_v0	get_rlimit_mask;

	/* from sched_cmd */
	struct vcmd_set_sched_v2	set_sched_v2;
	struct vcmd_set_sched_v3	set_sched_v3;
	struct vcmd_set_sched_v4	set_sched_v4;
	struct vcmd_sched_v5		set_sched_v5;
	struct vcmd_sched_v5		get_sched_v5;

	/* from signal_cmd */
	struct vcmd_ctx_kill_v0		ctx_kill;
	struct vcmd_wait_exit_v0	wait_exit;

	/* from space_cmd */
	struct vcmd_space_mask		enter_space;
	struct vcmd_space_mask		set_space;
	struct vcmd_space_mask		get_space_mask;

	/* from device_cmd */
	struct vcmd_set_mapping_v0	set_mapping;

	/* from debug and monitor */
	struct vcmd_read_history_v0	read_history;
	struct vcmd_read_monitor_v0	read_monitor;

};



int	handle_data(union cmd_arg *data, const char *str,
	uint32_t cmd, int index, int out)
{
#ifdef	DEBUG
	printf(" hadle_data(%p,%s,%x,%d,%d)\n",
	    data, str, cmd, index, out);
#endif
	CMDE_X(get_version		);
	CMDE_X(get_vci			);

	/* from legacy compat */

	CMDE_T(new_s_context,_v1,	0,1, uint32_t,	remove_cap);
	CMDE_T(new_s_context,_v1,	1,1, uint32_t,	flags);

	CMDE_T(set_ipv4root,_v3,	0,1, uint32_t,	broadcast);
	CMDE_T(set_ipv4root,_v3,	1,1, uint32_t,	nx_mask_pair[0].ip);
	CMDE_T(set_ipv4root,_v3,	2,1, uint32_t,	nx_mask_pair[0].mask);
	/* only one ip implemented */

	CMDE_X(create_context		);


	/* from context_cmd */

	CMDE_T(vx_info,,		0,2, uint32_t,	xid);
	CMDE_T(vx_info,,		1,2, uint32_t,	initpid);
	CMDE_X(task_xid			);

	CMDE_X(ctx_create_v0		);
	CMDE_T(ctx_create,,		0,1, uint64_t,	flagword);

	CMDE_X(ctx_create_v0		);
	CMDE_T(ctx_migrate,,		0,1, uint64_t,	flagword);

	CMDE_T(get_cflags,,		0,2, uint64_t,	flagword);
	CMDE_T(get_cflags,,		1,2, uint64_t,	mask);

	CMDE_T(set_cflags,,		0,3, uint64_t,	flagword);
	CMDE_T(set_cflags,,		1,3, uint64_t,	mask);

	CMDE_T(get_ccaps_v0,,		0,2, uint64_t,	bcaps);
	CMDE_T(get_ccaps_v0,,		1,2, uint64_t,	ccaps);
	CMDE_T(get_ccaps_v0,,		2,2, uint64_t,	cmask);

	CMDE_T(set_ccaps_v0,,		0,1, uint64_t,	bcaps);
	CMDE_T(set_ccaps_v0,,		1,1, uint64_t,	ccaps);
	CMDE_T(set_ccaps_v0,,		2,1, uint64_t,	cmask);

	CMDE_T(get_ccaps,_v1,		0,2, uint64_t,	ccaps);
	CMDE_T(get_ccaps,_v1,		1,2, uint64_t,	cmask);

	CMDE_T(set_ccaps,_v1,		0,1, uint64_t,	ccaps);
	CMDE_T(set_ccaps,_v1,		1,1, uint64_t,	cmask);

	CMDE_T(get_bcaps,,		0,2, uint64_t,	bcaps);
	CMDE_T(get_bcaps,,		1,2, uint64_t,	bmask);

	CMDE_T(set_bcaps,,		0,1, uint64_t,	bcaps);
	CMDE_T(set_bcaps,,		1,1, uint64_t,	bmask);


	/* from network_cmd */

	CMDE_T(nx_info,,		0,2, uint32_t,	nid);
	CMDE_X(task_nid			);

	CMDE_X(net_create_v0		);
	CMDE_T(net_create,,		0,1, uint64_t,	flagword);

	CMDE_X(net_migrate		);

	CMDE_T(net_add,,		0,1, uint16_t,	type);
	CMDE_T(net_add,,		1,1, uint16_t,	count);
	CMDE_N(net_add,,		2,1, in_addr,	ip[0]);
	CMDE_N(net_add,,		3,1, in_addr,	mask[0]);
	CMDE_N(net_add,,		4,1, in_addr,	ip[1]);
	CMDE_N(net_add,,		5,1, in_addr,	mask[1]);
	CMDE_N(net_add,,		6,1, in_addr,	ip[2]);
	CMDE_N(net_add,,		7,1, in_addr,	mask[2]);
	CMDE_N(net_add,,		8,1, in_addr,	ip[3]);
	CMDE_N(net_add,,		9,1, in_addr,	mask[3]);

	CMDE_T(net_remove,,		0,1, uint16_t,	type);
	CMDE_T(net_remove,,		1,1, uint16_t,	count);
	CMDE_N(net_remove,,		2,1, in_addr,	ip[0]);
	CMDE_N(net_remove,,		3,1, in_addr,	mask[0]);
	CMDE_N(net_remove,,		4,1, in_addr,	ip[1]);
	CMDE_N(net_remove,,		5,1, in_addr,	mask[1]);
	CMDE_N(net_remove,,		6,1, in_addr,	ip[2]);
	CMDE_N(net_remove,,		7,1, in_addr,	mask[2]);
	CMDE_N(net_remove,,		8,1, in_addr,	ip[3]);
	CMDE_N(net_remove,,		9,1, in_addr,	mask[3]);

	CMDE_T(net_add_ipv4,,		0,1, uint16_t,	type);
	CMDE_T(net_add_ipv4,,		1,1, uint16_t,	flags);
	CMDE_N(net_add_ipv4,,		2,1, in_addr,	ip);
	CMDE_N(net_add_ipv4,,		3,1, in_addr,	mask);

	CMDE_T(net_remove_ipv4,,	0,1, uint16_t,	type);
	CMDE_T(net_remove_ipv4,,	1,1, uint16_t,	flags);
	CMDE_N(net_remove_ipv4,,	2,1, in_addr,	ip);
	CMDE_N(net_remove_ipv4,,	3,1, in_addr,	mask);

	CMDE_T(net_add_ipv6,,		0,1, uint16_t,	type);
	CMDE_T(net_add_ipv6,,		1,1, uint16_t,	flags);
	CMDE_T(net_add_ipv6,,		2,1, uint32_t,	prefix);
	CMDE_N(net_add_ipv6,,		3,1, in6_addr,	ip);
	CMDE_N(net_add_ipv6,,		4,1, in6_addr,	mask);

	CMDE_T(net_remove_ipv6,,	0,1, uint16_t,	type);
	CMDE_T(net_remove_ipv6,,	1,1, uint16_t,	flags);
	CMDE_T(net_remove_ipv6,,	2,1, uint32_t,	prefix);
	CMDE_N(net_remove_ipv6,,	3,1, in6_addr,	ip);
	CMDE_N(net_remove_ipv6,,	4,1, in6_addr,	mask);

	CMDE_T(get_nflags,,		0,2, uint64_t,	flagword);
	CMDE_T(get_nflags,,		1,2, uint64_t,	mask);

	CMDE_T(set_nflags,,		0,1, uint64_t,	flagword);
	CMDE_T(set_nflags,,		1,1, uint64_t,	mask);

	CMDE_T(get_ncaps,,		0,2, uint64_t,	ncaps);
	CMDE_T(get_ncaps,,		1,2, uint64_t,	cmask);

	CMDE_T(set_ncaps,,		0,1, uint64_t,	ncaps);
	CMDE_T(set_ncaps,,		1,1, uint64_t,	cmask);


	/* from tag_cmd */

	CMDE_X(task_tag			);
	CMDE_X(tag_migrate		);


	/* from cvirt_cmd */

	CMDE_T(set_vhi_name,,		0,1, uint32_t,	field);
	CMDE_S(set_vhi_name,,		1,1, 65,	name);

	CMDE_T(get_vhi_name,,		0,3, uint32_t,	field);
	CMDE_S(get_vhi_name,,		1,2, 65,	name);


	/* from dlimit_cmd */

	CMDE_P(add_dlimit,,		0,1,		name);
	CMDE_T(add_dlimit,,		1,1, uint32_t,	flags);

	CMDE_P(rem_dlimit,,		0,1,		name);
	CMDE_T(rem_dlimit,,		1,1, uint32_t,	flags);

	CMDE_P(set_dlimit,,		0,1,		name);
	CMDE_T(set_dlimit,,		1,1, uint32_t,	space_used);
	CMDE_T(set_dlimit,,		2,1, uint32_t,	space_total);
	CMDE_T(set_dlimit,,		3,1, uint32_t,	inodes_used);
	CMDE_T(set_dlimit,,		4,1, uint32_t,	inodes_total);
	CMDE_T(set_dlimit,,		5,1, uint32_t,	reserved);
	CMDE_T(set_dlimit,,		6,1, uint32_t,	flags);

	CMDE_P(get_dlimit,,		0,1,		name);
	CMDE_T(get_dlimit,,		1,2, uint32_t,	space_used);
	CMDE_T(get_dlimit,,		2,2, uint32_t,	space_total);
	CMDE_T(get_dlimit,,		3,2, uint32_t,	inodes_used);
	CMDE_T(get_dlimit,,		4,2, uint32_t,	inodes_total);
	CMDE_T(get_dlimit,,		5,2, uint32_t,	reserved);
	CMDE_T(get_dlimit,,		6,2, uint32_t,	flags);


	/* from inode_cmd */

	CMDE_T(get_iattr_v0,,		0,1, uint64_t,	ino);
	CMDE_T(get_iattr_v0,,		1,3, uint32_t,	xid);
	CMDE_T(get_iattr_v0,,		2,2, uint32_t,	flags);
	CMDE_T(get_iattr_v0,,		3,2, uint32_t,	mask);

	CMDE_T(set_iattr_v0,,		0,1, uint64_t,	ino);
	CMDE_T(set_iattr_v0,,		1,3, uint32_t,	xid);
	CMDE_T(set_iattr_v0,,		2,3, uint32_t,	flags);
	CMDE_T(set_iattr_v0,,		3,3, uint32_t,	mask);

	CMDE_P(get_iattr,_v1,		0,1,		name);
	CMDE_T(get_iattr,_v1,		1,3, uint32_t,	tag);
	CMDE_T(get_iattr,_v1,		2,2, uint32_t,	flags);
	CMDE_T(get_iattr,_v1,		3,2, uint32_t,	mask);

	CMDE_P(set_iattr,_v1,		0,1,		name);
	CMDE_T(set_iattr,_v1,		1,3, uint32_t,	tag);
	CMDE_T(set_iattr,_v1,		2,3, uint32_t,	flags);
	CMDE_T(set_iattr,_v1,		3,3, uint32_t,	mask);


	/* from rlimit_cmd */

	CMDE_T(get_rlimit,,		0,1, uint32_t,	id);
	CMDE_T(get_rlimit,,		1,2, uint64_t,	minimum);
	CMDE_T(get_rlimit,,		2,2, uint64_t,	softlimit);
	CMDE_T(get_rlimit,,		3,2, uint64_t,	maximum);

	CMDE_T(set_rlimit,,		0,1, uint32_t,	id);
	CMDE_T(set_rlimit,,		1,1, uint64_t,	minimum);
	CMDE_T(set_rlimit,,		2,1, uint64_t,	softlimit);
	CMDE_T(set_rlimit,,		3,1, uint64_t,	maximum);

	CMDE_T(get_rlimit_mask,,	0,2, uint32_t,	minimum);
	CMDE_T(get_rlimit_mask,,	1,2, uint32_t,	softlimit);
	CMDE_T(get_rlimit_mask,,	2,2, uint32_t,	maximum);

	CMDE_X(reset_hits		);

	CMDE_X(reset_minmax		);

	/* from sched_cmd */

	CMDE_T(set_sched_v2,,		0,1, int32_t,	fill_rate);
	CMDE_T(set_sched_v2,,		1,1, int32_t,	interval);
	CMDE_T(set_sched_v2,,		2,1, int32_t,	tokens);
	CMDE_T(set_sched_v2,,		3,1, int32_t,	tokens_min);
	CMDE_T(set_sched_v2,,		4,1, int32_t,	tokens_max);
	CMDE_T(set_sched_v2,,		5,1, uint64_t,	cpu_mask);

	CMDE_T(set_sched_v3,,		0,1, uint32_t,	set_mask);
	CMDE_T(set_sched_v3,,		1,1, int32_t,	fill_rate);
	CMDE_T(set_sched_v3,,		2,1, int32_t,	interval);
	CMDE_T(set_sched_v3,,		3,1, int32_t,	tokens);
	CMDE_T(set_sched_v3,,		4,1, int32_t,	tokens_min);
	CMDE_T(set_sched_v3,,		5,1, int32_t,	tokens_max);
	CMDE_T(set_sched_v3,,		6,1, int32_t,	priority_bias);

	CMDE_T(set_sched_v4,,		0,1, uint32_t,	set_mask);
	CMDE_T(set_sched_v4,,		1,1, int32_t,	fill_rate);
	CMDE_T(set_sched_v4,,		2,1, int32_t,	interval);
	CMDE_T(set_sched_v4,,		3,1, int32_t,	tokens);
	CMDE_T(set_sched_v4,,		4,1, int32_t,	tokens_min);
	CMDE_T(set_sched_v4,,		5,1, int32_t,	tokens_max);
	CMDE_T(set_sched_v4,,		6,1, int32_t,	prio_bias);
	CMDE_T(set_sched_v4,,		7,1, int32_t,	cpu_id);
	CMDE_T(set_sched_v4,,		8,1, int32_t,	bucket_id);

	CMDE_T(set_sched,_v5,		0,1, uint32_t,	mask);
	CMDE_T(set_sched,_v5,		1,1, int32_t,	cpu_id);
	CMDE_T(set_sched,_v5,		2,1, int32_t,	bucket_id);
	CMDE_T(set_sched,_v5,		3,1, int32_t,	fill_rate[0]);
	CMDE_T(set_sched,_v5,		4,1, int32_t,	interval[0]);
	CMDE_T(set_sched,_v5,		5,1, int32_t,	fill_rate[1]);
	CMDE_T(set_sched,_v5,		6,1, int32_t,	interval[1]);
	CMDE_T(set_sched,_v5,		7,1, int32_t,	tokens);
	CMDE_T(set_sched,_v5,		8,1, int32_t,	tokens_min);
	CMDE_T(set_sched,_v5,		9,1, int32_t,	tokens_max);
	CMDE_T(set_sched,_v5,		10,1,int32_t,	prio_bias);

	CMDE_T(get_sched,_v5,		0,3, uint32_t,	mask);
	CMDE_T(get_sched,_v5,		1,1, int32_t,	cpu_id);
	CMDE_T(get_sched,_v5,		2,1, int32_t,	bucket_id);
	CMDE_T(get_sched,_v5,		3,2, int32_t,	fill_rate[0]);
	CMDE_T(get_sched,_v5,		4,2, int32_t,	interval[0]);
	CMDE_T(get_sched,_v5,		5,2, int32_t,	fill_rate[1]);
	CMDE_T(get_sched,_v5,		6,2, int32_t,	interval[1]);
	CMDE_T(get_sched,_v5,		7,2, int32_t,	tokens);
	CMDE_T(get_sched,_v5,		8,2, int32_t,	tokens_min);
	CMDE_T(get_sched,_v5,		9,2, int32_t,	tokens_max);
	CMDE_T(get_sched,_v5,		10,2,int32_t,	prio_bias);


	/* from signal_cmd */

	CMDE_T(ctx_kill,,		0,1, int32_t,	pid);
	CMDE_T(ctx_kill,,		1,1, int32_t,	sig);

	CMDE_T(wait_exit,,		0,2, int32_t,	reboot_cmd);
	CMDE_T(wait_exit,,		1,2, int32_t,	exit_code);


	/* from space_cmd */

	CMDE_X(enter_space_v0		);
	CMDE_X(set_space_v0		);

	CMDE_T(enter_space,,		0,1, uint64_t,	mask);
	CMDE_T(set_space,,		0,1, uint64_t,  mask);

	CMDE_T(get_space_mask,,		0,2, uint64_t,  mask);


	/* from device_cmd */

	CMDE_P(set_mapping,,		0,1,		device);
	CMDE_P(set_mapping,,		1,1,		target);
	CMDE_T(set_mapping,,		2,1, int32_t,	flags);


	/* debug and monitor cmds */

	CMDE_X(dump_history		);
	CMDE_T(read_history,,	0,3, uint32_t,	index);
	CMDE_T(read_history,,	1,3, uint32_t,	count);
	// CMDE_P(read_history,,	2,3,		data);
	CMDE_T(read_monitor,,	0,3, uint32_t,	index);
	CMDE_T(read_monitor,,	1,3, uint32_t,	count);
	// CMDE_P(read_monitor,,	2,3,		data);

#ifdef	DEBUG
	printf(" hadle_data(%p,%s,%x,%d,%d) ret=-1\n",
	    data, str, cmd, index, out);
#endif

	return -1;
}


int	set_index(union cmd_arg *data, uint32_t cmd, int index, char *str)
{
	int ret;
#ifdef	DEBUG
	printf(" set index %d = %s\n", index, str);
#endif
	ret = handle_data(data, str, cmd, index, 0);
	if (ret == -1)
	    return -1;
	return ++index;
}

int	set_named(union cmd_arg *data, uint32_t cmd, char *str)
{
	char *pos = strchr(str, '=');
	int index;

	if (!pos)
	    return -1;

	*pos = '\0';
	index = handle_data(data, str, cmd, 0, -1);
	*pos = '=';
	if (index == -1)
	    return -1; 
#ifdef	DEBUG
	printf(" index of %s = %d\n", str, index);
#endif
	return set_index(data, cmd, index, pos+1);
}

int	parse_cmd(char *str)
{
	int cmd = 0;

	if (strncmp(str, "VCMD_", 5) == 0)
	    str += 5;
	if (str[0] == '#')
	    cmd = argtoll(str + 1,0);
	else
	    cmd = handle_data(NULL, str, 0, -2, -2);
	if (cmd == -1) {
	    int c = (unsigned)argtoll(strtok(str, "_"),0);
	    int i = (unsigned)argtoll(strtok(NULL, "["),0);
	    int v = (unsigned)argtoll(strtok(NULL, "]"),0);

	    if (v == -1)
		v = 0;
	    if ((c != -1) && (i != -1))
		cmd = (((c & 0x3F) << 24) |
		    ((i & 0xFF) << 16) | (v & 0xFFF));
	}
#ifdef	DEBUG
	printf(" cmd from %s = %x\n", str, cmd);
#endif
	return cmd;
}

int	get_io(uint32_t cmd, int index)
{
	return handle_data(NULL, "", cmd, index, -2);
}

int	has_data(uint32_t cmd)
{
	return handle_data(NULL, "", cmd, -1, -2);
}


#define	CMD_FMT		"VCMD_%02d_%d[%d]"
#define	CMD_ARG(c)	VC_CATEGORY(c), VC_COMMAND(c), VC_VERSION(c)

void	dump_data(union cmd_arg *data, int out)
{
	int index = 0;
	int io;

	printf(CMD_FMT " {\n", CMD_ARG(num_cmd));
	index = 0;
	do {
	    int r = 0;

	    io = get_io(num_cmd, index);
	    if (opt_terse &&
		((io & out) == 0))
		goto skip;

	    r = handle_data(data, NULL,
		num_cmd, index, 1);
	    if (r < 0)
		break;
	skip:
	    index++;
	} while (io);
	printf("}\n");
}

void	dump_info(void)
{
	union cmd_arg data = { .info.cmd_name = "unknown" };
	int r;

	r = handle_data(&data, NULL, num_cmd, -1, -4);
	if (r == 1)
	    printf("VCMD_%s [v%d]:\n", 
		data.info.cmd_name, VC_VERSION(num_cmd));
	else
	    printf("unknown VCMD:\n");
}

void	dump_desc(void)
{
	int index = 0;
	int r;

	printf(CMD_FMT " {\n", CMD_ARG(num_cmd));
	do {
	    r = handle_data(NULL, NULL,
		num_cmd, index, -3);
	    index++;
	} while ((r >= 0));
	printf("}\n");
}

void	do_syscall(union cmd_arg *data)
{
	num_ret = vserver(num_cmd, num_id, data);
}

struct err_data *get_err_data(int error)
{
	struct err_data *ptr;

	for (ptr=err_data; ptr->code >= 0; ptr++)
	    if (ptr->code == error)
		break;

	return ptr;
}


void do_exec(char *cmd, char *argv[])
{
	int ret = 0;

	if (opt_clone) {
	    pid_t pid;
	    int status;

	    if ((pid = clone(flags_clone, NULL))) {
		/* parent */		
		if (ret<0) {
		    perror("clone");
		    exit(6);
		} else {
		    waitpid(-1, &status, __WCLONE);
		    exit(status);
		}
	    } 
	}

	if (opt_unshare) {
	    ret = unshare(flags_unshare);
	    if (ret<0)
	        perror("unshare");
	}
	
	exec_name = cmd;
	ret = execvp(exec_name, argv);
	
	if (ret<0)
	    perror("execvp");
	exit(5);
}


int capget(cap_user_header_t hdrp, cap_user_data_t datap);
int capset(cap_user_header_t hdrp, const cap_user_data_t datap);


#define OPTIONS "hc:di:klnqtu:ABC:L:M:V"

int	main(int argc, char *argv[])
{
	extern int optind;
	extern char *optarg;
	char c, errflg = 0;
	char dataflg;
	int r, index;
	union cmd_arg vc_data;

	cmd_name = argv[0];
	while ((c = getopt(argc, argv, "+" OPTIONS)) != EOF) {
	    switch (c) {
	    case 'h':
		fprintf(stderr,
		    "This is %s " VERSION "\n"
		    "options are:\n"
		    "  -h        print this help message\n"
		    "  -A        dump data after syscall\n"
		    "  -B        dump data before syscall\n"
		    "  -c <flgs> use clone before exec\n"
		    "  -d        describe command data\n"
		    "  -i <num>  id value\n"
		    "  -k        continue after error\n"
		    "  -l        list command and args\n"
		    "  -n        dry run without syscall\n"
		    "  -q        quiet, returncode only\n"
		    "  -t        terse, show only relevant\n"
		    "  -u <flgs> unshare before exec\n"
		    "  -C <cmd>  vcmd command name\n"
		    "  -L <caps> linux capset\n"
		    "  -M <cap>  posix capability mask\n"
		    "  -V        verify interface version\n"
		    "  --        end of options\n"
		    ,cmd_name);
		exit(0);
		break;
	    case 'c':
		opt_clone = 1;
		flags_clone = ato_uint64_t(optarg);
		break;
	    case 'd':
		opt_desc = 1;
		break;
	    case 'i':
		num_id = ato_uint32_t(optarg);
		opt_id = 1;
		break;
	    case 'k':
		opt_cont = 1;
		break;
	    case 'l':
		opt_list = 1;
		break;
	    case 'n':
		opt_dryrun = 1;
		break;
	    case 'q':
		opt_quiet = 1;
		break;
	    case 't':
		opt_terse = 1;
		break;
	    case 'u':
		opt_unshare = 1;
		flags_unshare = ato_uint64_t(optarg);
		break;
	    case 'A':
		opt_after = 1;
		break;
	    case 'B':
		opt_before = 1;
		break;
	    case 'C':
		num_cmd = parse_cmd(optarg);
		if (num_cmd == -1) {
		    printf("unrecognized vc %s, aborting.\n",
			optarg);
		    exit(3);
		}
		opt_cmd = 1;
		break;
	    case 'L':
		parse_caps(optarg);
		opt_cap = 1;
		break;
	    case 'M':
		num_mask = argtoll(optarg, stobcap);
		opt_mask = 1;
		break;
	    case 'V':
		opt_vers = 1;
		break;
	    case '?':
	    default:
		errflg++;
		break;
	    }
	}
	if (errflg) {
	    fprintf(stderr, 
		"Usage: %s [-" OPTIONS "] <args> [-e cmd ...]\n"
		"%s -h for help.\n",
		cmd_name, cmd_name);
	    exit(2);
	}

	if (opt_vers) {
	    r = vserver(VCMD_get_version, 0, NULL);
	    if (r<0)
		perror("vc_get_version");
	    else 
		printf("version: %04x:%04x\n", 
		    (r>>16) & 0xFFFF, r & 0xFFFF);
	}

	memset(&vc_data, 0, sizeof(vc_data));
	dataflg = has_data(num_cmd);

	if (opt_desc) {
	    dump_info();
	    if (!dataflg)
		printf(CMD_FMT " does not take arguments.\n",
		    CMD_ARG(num_cmd));
	    else if (!opt_terse)
		dump_desc();
	}

	if (!dataflg)
	    goto skip_data;
	if ((optind > 0) && strcmp(argv[optind-1], "--") == 0)
	    goto skip_data;

	index = 0;
	for (; optind < argc; optind++) {
	    if (strcmp(argv[optind], "--") == 0) {
		optind++;
		break;
	    }

	    /* skip to next input field */
	    while (!(get_io(num_cmd, index) & 1))
		index++;

	    /* handle named data */
	    if (argv[optind][0] == '.')
		index = set_named(&vc_data,
		    num_cmd, argv[optind]+1);
	    else
		index = set_index(&vc_data,
		    num_cmd, index, argv[optind]);
	    if (index == -1)
		printf("warning: argument %d %s ignored.\n",
		    optind, argv[optind]);
	}


	if (opt_before)
	    dump_data(&vc_data, 1);

skip_data:
	if (opt_dryrun) 
	    goto skip_syscall;

	do_syscall(&vc_data);

	if (num_ret == -1) {
	    num_err = errno;
	    if (opt_desc) {
		struct err_data *ed = get_err_data(num_err);
		
		printf("vserver: %s [%d] %s\n",
		    ed->symb, num_err, ed->desc);
	    }
	    else if (!opt_quiet)
		perror("vserver");
	    if (!opt_cont)
		exit(-num_err);
	} else if (!opt_quiet) {
	    printf("vserver: ret = " FMT_int32_t "\n",
		(long long)num_ret, (long long)num_ret);
	}

	if (opt_after && dataflg)
	    dump_data(&vc_data, 2);

skip_syscall:
	if (opt_cap) {
		int ret, pid = getpid();
		struct __user_cap_header_struct hdr = { _LINUX_CAPABILITY_VERSION, pid };
		struct __user_cap_data_struct data = { cap_eff, cap_prm, cap_inh };
		struct err_data *ed;

		ret = capset(&hdr, &data);
	    	if (opt_desc) {
		    ed = get_err_data(errno);
		    printf("capset: %08x:%08x:%08x = %s [%d] %s\n",
			cap_eff, cap_prm, cap_inh, ed->symb, errno, ed->desc);
		}
		else if (ret<0)
		    perror("capset");
		ret = capget(&hdr, &data);
	    	if (opt_desc) {
		    ed = get_err_data(errno);
		    printf("capget: %08x:%08x:%08x = %s [%d] %s\n",
		        data.effective, data.permitted, data.inheritable,
			ed->symb, errno, ed->desc);
		}
		else if (ret<0)
		    perror("capget");
	}
	if (opt_list) {
	    for (index=0; optind < argc; optind++, index++)
		printf("cmd[%d] = %s\n", index, argv[optind]);
	    optind -= index;
	}
	if (optind < argc) {
	    do_exec(argv[optind], argv+optind);
	    exit(6);
	}

	exit(num_ret);
}

